﻿#include <windows.h>
#include <string>

extern "C" { 
        #include <libavcodec/avcodec.h> 
        #include <libavformat/avformat.h> 
        #include <libswscale/swscale.h> 

}
#if _MSC_VER
	#pragma comment(lib, "libavcodec.lib")
	#pragma comment(lib, "libavformat.lib")
	#pragma comment(lib, "libavutil.lib")
	#pragma comment(lib, "libswscale.lib")


	#pragma comment(lib, "zlib.lib")
	#pragma comment(lib, "ws2_32.lib")
#endif

bool SaveFrame(int nszBuffer, uint8_t *buffer, char cOutFileName[])
{
	bool bRet = false;
	FILE *pFile;
	if (nszBuffer > 0)
	{
		if (fopen_s(&pFile, cOutFileName, "wb") == 0)
		{
			fwrite(buffer, sizeof(uint8_t), nszBuffer, pFile);
			bRet = true;
			fclose(pFile);
		}
	}
	return bRet;
}

bool WriteJPEG(AVCodecContext *pCodecCtx, AVFrame *pFrame, char cFileName[], PixelFormat pix, uint8_t *buffer, int numBytes)
{
	bool bRet = false;
	AVCodec *pMJPEGCodec = NULL;
	AVCodecContext *pMJPEGCtx = avcodec_alloc_context3(NULL);
	if (pMJPEGCtx)
	{
		pMJPEGCtx->bit_rate = pCodecCtx->bit_rate;
		pMJPEGCtx->width = pCodecCtx->width;
		pMJPEGCtx->height = pCodecCtx->height;
		pMJPEGCtx->pix_fmt = pix;
		pMJPEGCtx->codec_id = CODEC_ID_MJPEG;
		pMJPEGCtx->codec_type = AVMEDIA_TYPE_VIDEO;
		pMJPEGCtx->time_base.num = pCodecCtx->time_base.num;
		pMJPEGCtx->time_base.den = pCodecCtx->time_base.den;
		pMJPEGCodec = avcodec_find_encoder(pMJPEGCtx->codec_id);

		if (pMJPEGCodec && (avcodec_open2(pMJPEGCtx, pMJPEGCodec, NULL) >= 0))
		{
			pMJPEGCtx->qmin = pMJPEGCtx->qmax = 3;
			pMJPEGCtx->mb_lmin = pMJPEGCtx->lmin = pMJPEGCtx->qmin * FF_QP2LAMBDA;
			pMJPEGCtx->mb_lmax = pMJPEGCtx->lmax = pMJPEGCtx->qmax * FF_QP2LAMBDA;
			pMJPEGCtx->flags |= CODEC_FLAG_QSCALE;
			pFrame->quality = 10;
			pFrame->pts = 0;
			int szBufferActual = avcodec_encode_video(pMJPEGCtx, buffer, numBytes, pFrame);

			if (SaveFrame(szBufferActual, buffer, cFileName))
				bRet = true;
			avcodec_close(pMJPEGCtx);
		}
	}
	return bRet;
}

bool PullFrame(char cInPath[], char cFileName[], char cImgOutPath[], char cImgOutName[], int nTimeStamp)
{
	bool bRet = false;
	int videoStream = -1;
	AVFrame *pFrame = NULL;
	AVFrame *pFrameRGB = NULL;
	AVPacket packet;
	int frameFinished = 0;

	AVDictionary *optionsDict = NULL;

	char cFileLocale[MAX_PATH];
	sprintf_s(cFileLocale, MAX_PATH, "%s%s", cInPath, cFileName);

	avformat_network_init();
	av_register_all();

	AVFormatContext *pFormatCtx = avformat_alloc_context();
	if (avformat_open_input(&pFormatCtx, cFileLocale, NULL, NULL) == 0)
	{
		if (avformat_find_stream_info(pFormatCtx, NULL) >= 0)
		{
			av_dump_format(pFormatCtx, 0, cFileName, 0);
			for (int i = 0; i<(int)pFormatCtx->nb_streams; i++)
			{
				if (pFormatCtx->streams[i]->codec->codec_type == AVMEDIA_TYPE_VIDEO)
				{
					videoStream = i;
					break;
				}
			}
			if (videoStream >= 0)
			{
				AVCodecContext *pCodecCtx = pFormatCtx->streams[videoStream]->codec;
				AVCodec *pCodec = avcodec_find_decoder(pCodecCtx->codec_id);
				if (pCodec != NULL)
				{
					if (avcodec_open2(pCodecCtx, pCodec, &optionsDict) >= 0)
					{
						pFrame = avcodec_alloc_frame();
						pFrameRGB = avcodec_alloc_frame();
						if (pFrameRGB != NULL)
						{
							int nFramesSaved = 0;
							int numBytes = avpicture_get_size(PIX_FMT_YUVJ420P, pCodecCtx->width, pCodecCtx->height);
							uint8_t *buffer = (uint8_t *)av_malloc(numBytes*sizeof(uint8_t));

							avpicture_fill((AVPicture *)pFrameRGB, buffer, PIX_FMT_YUVJ420P, pCodecCtx->width, pCodecCtx->height);
							nTimeStamp = nTimeStamp * AV_TIME_BASE;
							int nRet = av_seek_frame(pFormatCtx, -1, nTimeStamp + pFormatCtx->start_time, AVSEEK_FLAG_ANY);
							if (nRet < 0)
								nRet = av_seek_frame(pFormatCtx, -1, nTimeStamp + pFormatCtx->start_time, AVSEEK_FLAG_BACKWARD);
							if (nRet >= 0)
							{
								int i = 0;
								while (av_read_frame(pFormatCtx, &packet) >= 0)
								{
									if (packet.stream_index == videoStream)
									{
										avcodec_decode_video2(pCodecCtx, pFrame, &frameFinished, &packet);
										if (frameFinished)
										{
											if (++i <= 5)
											{
												char cFileName[MAX_PATH];
												sprintf_s(cFileName, MAX_PATH, "%sframe%d.jpg", cImgOutPath, i);
												WriteJPEG(pCodecCtx, pFrame, cFileName, PIX_FMT_YUVJ420P, buffer, numBytes);
											}
											else
											{
												av_free_packet(&packet);
												break;
											}
										}
									}
									av_free_packet(&packet);
								}
							}
						}
						else
							bRet = false;

						av_free(pFrameRGB);
						av_free(pFrame);
						avcodec_close(pCodecCtx);
					}
					else
						bRet = false;
				}
				else
					bRet = false;
			}
			else
				bRet = false;
		}
		else
			bRet = false;
		avformat_close_input(&pFormatCtx);
	}
	return bRet;
}


#define CHECK_ERR(ERR) {if ((ERR)<0) return -1; }

int convert_first_frame_to_png(std::string const & inputVideoFileName, std::string const & outputPngName)
{
	av_register_all();
	avcodec_register_all();

	::AVFormatContext * ctx = NULL;
	int err = avformat_open_input(&ctx, inputVideoFileName.c_str(), NULL, NULL);
	CHECK_ERR(err);
	err = avformat_find_stream_info(ctx, NULL);
	CHECK_ERR(err);

	AVCodec * codec = NULL;
	int strm = av_find_best_stream(ctx, AVMEDIA_TYPE_VIDEO, -1, -1, &codec, 0);

	AVCodecContext * codecCtx = ctx->streams[strm]->codec;
	err = avcodec_open2(codecCtx, codec, NULL);
	CHECK_ERR(err);
	if (avformat_seek_file(ctx, -1, INT64_MIN, 10, INT64_MAX, 0) < 0)
	{
		return -1; // can not seek
	}

	SwsContext * swCtx = sws_getContext(codecCtx->width,
		codecCtx->height,
		codecCtx->pix_fmt,
		codecCtx->width,
		codecCtx->height,
		PIX_FMT_RGB24,
		SWS_FAST_BILINEAR, 0, 0, 0);

	for (;;)
	{
		AVPacket pkt;
		err = av_read_frame(ctx, &pkt);
		CHECK_ERR(err);

		if (pkt.stream_index == strm)
		{
			int got = 0;
			AVFrame * frame = avcodec_alloc_frame();
			err = avcodec_decode_video2(codecCtx, frame, &got, &pkt);
			CHECK_ERR(err);

			if (got)
			{
				AVFrame * rgbFrame = avcodec_alloc_frame();
				avpicture_alloc((AVPicture *)rgbFrame, PIX_FMT_RGB24, codecCtx->width, codecCtx->height);
				sws_scale(swCtx, frame->data, frame->linesize, 0, frame->height, rgbFrame->data, rgbFrame->linesize);

				AVCodec *outCodec = avcodec_find_encoder(CODEC_ID_PNG);
				if (!outCodec)
				{
					return -1;
				}
				AVCodecContext *outCodecCtx = avcodec_alloc_context3(outCodec);
				if (!codecCtx)
					return -1;

				outCodecCtx->width = codecCtx->width;
				outCodecCtx->height = codecCtx->height;
				outCodecCtx->pix_fmt = PIX_FMT_RGB24;
				outCodecCtx->codec_type = AVMEDIA_TYPE_VIDEO;
				outCodecCtx->time_base.num = codecCtx->time_base.num;
				outCodecCtx->time_base.den = codecCtx->time_base.den;

				if (avcodec_open2(outCodecCtx, outCodec, NULL) < 0) {
					return -1;
				}

				AVPacket outPacket;
				av_init_packet(&outPacket);
				outPacket.size = 0;
				outPacket.data = NULL;
				int gotFrame = 0;
				int ret = avcodec_encode_video2(outCodecCtx, &outPacket, rgbFrame, &gotFrame);
				if (ret >= 0 && gotFrame)
				{
					FILE * outPng = fopen(outputPngName.c_str(), "wb");
					fwrite(outPacket.data, outPacket.size, 1, outPng);
					fclose(outPng);
				}

				avcodec_close(outCodecCtx);
				av_free(outCodecCtx);

				break;
			}
			avcodec_free_frame(&frame);
		}
	}
	return 1;
}




#if _MSC_VER
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPreInst,
                   LPSTR lpszCmdLine, int nCmdShow)
#else
int main(int argc, const char **argv)
#endif
{
	PullFrame(".\\","demo_douga.avi", ".\\","foo",100);

	return 0;
}
